#    Copyright 2019-2020 Matthew Wigginton Conway

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# Author: Matthew Wigginton Conway <matt@indicatrix.org>
#         School of Geographical Sciences and Urban Planning
#         Arizona State University

from logging import getLogger
import time
import statsmodels.api as sm
import pandas as pd
import numpy as np
import datetime
import multiprocessing
import threading
import queue

from . import price_income
from .clear_market import ClearMarket
from eqsormo.common import BaseSortingModel, MNLFullASC
from eqsormo.common.util import human_bytes, human_time, human_shape, ConvergenceError
from eqsormo.common.compute_ascs import compute_ascs
import eqsormo
from . import save_load

from typing import Dict, Callable, Any

LOG = getLogger(__name__)


class TraSortingModel(BaseSortingModel):
    """
    The variety of sorting model described in Tra CI (2013) Measuring the General Equilibrium Benefits of Air Quality
    Regulation in Small Urban Areas. Land Economics 89(2): 291–307.

    The key difference between this and the Bayer et al. model is that price is included in the first stage rather than
    the second stage. This obviates the need to instrument for price, because there are housing type fixed effects in
    the first stage. But we can't include price directly, due to identification challenges since the prices could also
    be captured by the ASCs. So instead we include a budget constraint of the form f(income, price) - Tra uses
    ln(income - price) but the exact functional form is not important, as long as income (or another household
    attribute) is somehow interacted with price so that there is variation in price between households within a single
    housing type.

    The other key difference from both the Bayer and the Tra formulations is that this model supports "unequilibrated
    choice," that is it can model a joint choice between a good for which prices are adjusted to bring the market to
    equilibrium and one for which they are not. In my dissertation I use this to model the joint choice of housing
    (equilibrated good) and vehicle ownership (unequilibrated good).
    """

    def __init__(
        self,
        housing_attributes,
        household_attributes,
        interactions,
        unequilibrated_hh_params,
        unequilibrated_hsg_params,
        second_stage_params,
        price,
        income,
        choice,
        unequilibrated_choice,
        fixed_price,
        price_income_transformation=price_income.logdiff,
        price_income_starting_values=[],
        endogenous_variable_defs: Dict[str, Callable[[Any, Any], float]] = None,
        neighborhoods=None,
        sample_alternatives=None,
        method="L-BFGS-B",
        minimize_options={},
        max_rent_to_income=None,
        household_housing_attributes=None,
        weights=None,
        max_chunk_bytes=2e9,
        est_first_stage_ses=True,
        seed=None,
    ):
        """
        Initialize a Tra sorting model

        :param housing_attributes: Attributes of housing choices. Price should not be an attribute here.
        :type housing_attributes: Pandas dataframe.

        :param household_attributes: Attributes of households. It is okay if income is an attribute here as long as it
            is also passed in separately - since we assume it doesn't change when a household moves.
        :type household_attributes: Pandas dataframe

        :param household_housing_attributes: Attributes of the combination of households and housing choices,
            multiindexed with household ID and choice ID. Must contain all possible combinations.
        :type household_housing_attributes: Pandas dataframe

        :param interactions: Which household attributes should be interacted with which housing attributes in the
            first stage. Endogenous variables may be included as housing attributes here.
        :type interactions: iterable of tuple (household_attributes column, housing_attributes column)

        :param second_stage_params: Which housing attributes to include in the second stage fit (decomposition of mean
            parameters). If None, don't fit a second stage (this stage is not necessary for simulation). Note that
            currently endogenous variables cannot be used in this stage.
        :type second_stage_params: iterable of string

        :param price: Observed price of each housing choice. This price should be over the same time period as income
            (for example, annualized prices with annual incomes)
        :type price: Pandas series of numbers, indexed like housing_attributes

        :param income: Household income
        :type income: Pandas series of numbers, indexed like household_attributes

        :param choice: The choice made by the household. Should be index values in household_attributes
        :type choice: Pandas series, indexed like household_attributes

        :param fixed_price: Which price to fix in price finding
        :type fixed_price: float


        :param endogenous_variable_defs: This is a dictionary mapping variable names to functions that calculate them,
            for endogenous variables that need to be updated in the sorting stage of the model, as household re-sort
            themselves across the region. The function should accept a data frame of housing attributes, a series with
            household income, and a numpy array of weights for how likely that household is to live in a particular
            neighborhood. Should return a float scalar for that neighborhood.
        :type endogenous_variable_defs: Dict of string -> function

        :param neighborhoods: Pandas series of which neighborhood each housing choice is in, for calculation of
            endogenous variables. Indexed like housing_attributes.
        :type neighborhoods: pd.Series

        :param sample_alternatives: If a positive integer, take a random sample of this size for available alternatives
            to estimate the model. Recommended in models with a large number of alternatives. If 0 or None, no sampling
                will be performed.
        :type sample_alternatives: int or None

        :param method: A scipy.optimize.minimize method used for maximizing log-likelihood of first stage, default
            'L-BFGS-B'
        :type method: str

        :param minimize_options: Options for selected minimizer
        :type minimize_options: dict


        :param max_rent_to_income: Maximum proportion of income rent can be for the alternative to be included in the
            choice set - in (0, 1] if the price/income transformation can't handle income less than rent
        :type max_rent_to_income: float or None

        :param max_chunk_bytes: Maximum number of bytes to use for a single chunk of the full alternatives array.
            The most memory-constrained part of the alogorithm is clearing the market, which requires computing the dot
            product of a n_households x n_housing_alternatives * n_unequilibrated_alternatives array and a coefficients
            array. However, since the coefficients array is a column vector, we can compute the product of chunks of the
            full alternatives array and the column vector, then concatenate.
            This tuning parameter controls how large these chunks are in bytes. Default 2GB.
        :type max_chunk_bytes: int

        :param est_first_stage_ses: If False, do not estimate standard errors for the first stage. This can
            significantly speed convergence and is recommended during model development.
        :type est_first_stage_ses: bool



        :param seed: seed for the random number generator used in sampling alternatives
        :type seed: int
        """
        self.housing_attributes = housing_attributes
        self.household_attributes = household_attributes
        self.household_housing_attributes = household_housing_attributes
        self.interactions = interactions
        self.second_stage_params = second_stage_params
        self.price = price.reindex(housing_attributes.index)
        self.orig_price = self.price.copy()
        self.income = income.reindex(household_attributes.index)
        self.choice = choice.reindex(household_attributes.index)
        self.unequilibrated_choice = unequilibrated_choice.reindex(
            household_attributes.index
        )
        self.unequilibrated_hh_params = unequilibrated_hh_params
        self.unequilibrated_hsg_params = unequilibrated_hsg_params
        self.sample_alternatives = sample_alternatives
        self.alternatives_stds = None

        if weights is None:
            self.weights = None
            self.weighted_supply = self.choice.value_counts()
        else:
            self.weights = weights.reindex(household_attributes.index)
            self.weighted_supply = self.weights.groupby(self.choice).sum()

        self.endogenous_variable_defs = endogenous_variable_defs
        if neighborhoods is not None:
            self.neighborhoods = neighborhoods.reindex(self.housing_attributes.index)
        else:
            self.neighborhoods = None

        self.price_income_transformation = price_income_transformation
        self.price_income_starting_values = price_income_starting_values
        self.method = method
        self.minimize_options = minimize_options
        self.max_rent_to_income = max_rent_to_income
        self.max_chunk_bytes = max_chunk_bytes
        self.est_first_stage_ses = est_first_stage_ses
        self.fixed_price = fixed_price

        self.seed = seed
        self._rng = np.random.default_rng(seed=seed)

        self.validate()

        self.creation_time = datetime.datetime.today()

        assert (
            price_income_transformation.n_params == 0
        ), "Parameterized price_income_transformations not currently supported"

    def validate(self):
        # TODO this could easily be moved to a new file
        allPassed = True

        choiceCount = self.choice.value_counts().reindex(
            self.housing_attributes.index, fill_value=0
        )
        if not np.all(choiceCount > 0):
            choiceList = " - " + "\n - ".join(choiceCount.index[choiceCount == 0])
            LOG.error(
                f"Some housing alternatives are not chosen by any households!\n{choiceList}"
            )
            allPassed = False

        if self.max_rent_to_income is not None and not np.all(
            self.income.values * self.max_rent_to_income
            > self.price.loc[self.choice].values
        ):
            LOG.error(
                "Some households pay more in rent than the max rent to income ratio"
            )
            allPassed = False

        for hhattr, hsgattr in self.interactions:
            if hhattr not in self.household_attributes.columns:
                LOG.error(
                    f"Attribute {hhattr} is used in interactions but is not in household_attributes"
                )
                allPassed = False

            if self.household_attributes[hhattr].isnull().any():
                LOG.error(f"Attribute {hhattr} contains NaNs")
                allPassed = False

            if hsgattr not in self.housing_attributes.columns:
                if (
                    self.endogenous_variable_defs is None
                    or hsgattr not in self.endogenous_variable_defs
                ):
                    LOG.error(
                        f"Attribute {hsgattr} is used in interactions but is not in housing_attributes"
                    )
                    allPassed = False

            if (
                hsgattr in self.housing_attributes.columns
                and self.endogenous_variable_defs is not None
                and hsgattr in self.endogenous_variable_defs
            ):
                LOG.error(
                    f"{hsgattr} in both housing_attributes and endogenous variables"
                )
                allPassed = False

            if (
                hsgattr in self.housing_attributes.columns
                and self.housing_attributes[hsgattr].isnull().any()
            ):
                LOG.error(f"Attribute {hsgattr} contains NaNs")
                allPassed = False

        if self.second_stage_params is not None:
            for hsgattr in self.second_stage_params:
                if hsgattr not in self.housing_attributes.columns:
                    LOG.error(
                        f"Attribute {hsgattr} is used in second stage but is not in housing_attributes"
                    )
                    allPassed = False

                if self.housing_attributes[hsgattr].isnull().any():
                    LOG.error(f"Attribute {hsgattr} contains NaNs")
                    allPassed = False

        if self.endogenous_variable_defs is not None:
            if self.neighborhoods is None:
                LOG.error("Neighborhoods required when endogenous variables in use")
                allPassed = False
            if self.neighborhoods.isnull().any():
                LOG.error("All choices must have a neighborhood defined")
                allPassed = False

        # TODO more checks
        if allPassed:
            LOG.info("All validation checks passed!")
        else:
            raise ValueError("Some validation checks failed (see log messages)")

    def materialize_alternatives(
        self, hhidx, choiceidx, uneqchoiceidx, hh_hsgidx=None, price_income_params=None
    ):
        """
        Materialize the alternatives for hhidx, choiceidx, and uneqchoiceidx, and return them.

        These should be formatted like so, with hhidx changing slowest and uneqchoiceidx changing fastest.
        if there are three households, three housing choices, and three unequilibrated choices:
        hhidx:         0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2
        choiceidx:     0 0 0 1 1 1 2 2 2 0 0 0 1 1 1 2 2 2 0 0 0 1 1 1 2 2 2
        uneqchoiceidx: 0 1 2 0 1 2 0 1 2 0 1 2 0 1 2 0 1 2 0 1 2 0 1 2 0 1 2

        It is okay if they are not sequential, but they should be monotonically increasing.

        hh_hsgidx is the integer indices in household_housing_attributes for the selected household/housing combinations, same length
        as hhidx etc.
        """
        start_time = time.perf_counter()

        if price_income_params is None:
            # can't refer to self in function def
            price_income_params = self.price_income_starting_values

        assert len(hhidx) == len(choiceidx) and len(choiceidx) == len(uneqchoiceidx)

        if self.household_housing_attributes is not None:
            assert hh_hsgidx is not None and len(hhidx) == len(hh_hsgidx)

        LOG.debug(f"materializing {len(hhidx)} choices")

        # first, create data for the interactions
        colnames = []

        # + 1 for budget param
        ncols = (
            len(self.interactions)
            + (len(self.unequilibrated_hh_params) + len(self.unequilibrated_hsg_params))
            * (len(self.unequilibrated_choice_xwalk) - 1)
            + 1
        )

        if self.household_housing_attributes is not None:
            ncols += len(self.household_housing_attributes.columns)

        LOG.debug(
            f"Allocating alternatives array of size {human_bytes(len(hhidx) * ncols * 8)}"
        )
        alternatives = np.zeros((len(hhidx), ncols))

        # budget is first column, to make updates easier
        current_col = 0
        colnames.append("budget")
        alt_income = self.income.astype("float64").values[hhidx]
        alt_price = self.price.astype("float64").values[choiceidx]

        # don't calc buget for options not in choice set
        # it may throw an error (e.g. log(neg) for logdiff)
        if self.max_rent_to_income is not None:
            feasible_alts = alt_income * self.max_rent_to_income > alt_price
        else:
            feasible_alts = np.full(len(alt_income), True)

        LOG.debug(f"{np.sum(feasible_alts)} options appear in choice sets")

        budget = np.full(len(hhidx), np.nan)
        budget[feasible_alts] = self.price_income_transformation.apply(
            alt_income[feasible_alts], alt_price[feasible_alts], *price_income_params
        )
        assert not np.any(np.isnan(budget[feasible_alts]))  # should be no nans left
        alternatives[:, current_col] = budget
        current_col += 1
        del alt_income, alt_price, budget  # save memory

        for hh_attr, hsg_attr in self.interactions:
            if hsg_attr in self.housing_attributes.columns:
                # TODO lots of type conversion happening here. Could maybe refactor to do less.
                alternatives[:, current_col] = (
                    self.household_attributes[hh_attr].astype("float64").values[hhidx]
                    * self.housing_attributes[hsg_attr]
                    .astype("float64")
                    .values[choiceidx]
                )
            elif hsg_attr in self.endogenous_varnames:
                endogenous_col = self.endogenous_varnames.index(hsg_attr)
                alternatives[:, current_col] = (
                    self.household_attributes[hh_attr].astype("float64").values[hhidx]
                    * self.endogenous_variables[
                        self.nbhd_for_choice[choiceidx], endogenous_col
                    ]
                )
            else:
                raise KeyError(
                    f"{hsg_attr} is not a housing attribute, exogenous or endogenous"
                )
            colnames.append(f"{hh_attr}:{hsg_attr}")
            current_col += 1

        # now add the attributes for the unequilibrated choice
        for param in self.unequilibrated_hh_params:
            vals = self.household_attributes[param].astype("float64").values[hhidx]
            for uneqchoice in range(1, len(self.unequilibrated_choice_xwalk)):
                # fill all rows that are not for this unequilibrated choice with 0s
                alternatives[:, current_col] = np.choose(
                    uneqchoiceidx == uneqchoice, [0, vals]
                )
                colnames.append(
                    f"{param}:uneq_choice_{self.unequilibrated_choice_xwalk[self.unequilibrated_choice_xwalk == uneqchoice].index[0]}"
                )
                current_col += 1

        for param in self.unequilibrated_hsg_params:
            vals = self.housing_attributes[param].astype("float64").values[choiceidx]
            for uneqchoice in range(1, len(self.unequilibrated_choice_xwalk)):
                # fill all rows that are not for this unequilibrated choice with 0s
                alternatives[:, current_col] = np.choose(
                    uneqchoiceidx == uneqchoice, [0, vals]
                )
                colnames.append(
                    f"{param}:uneq_choice_{self.unequilibrated_choice_xwalk[self.unequilibrated_choice_xwalk == uneqchoice].index[0]}"
                )
                current_col += 1

        if self.household_housing_attributes is not None:
            for c in self.household_housing_attributes.columns:
                colnames.append(c)

            alternatives[
                :,
                current_col : current_col
                + len(self.household_housing_attributes.columns),
            ] = self.household_housing_attributes.values[hh_hsgidx, :]
            current_col += len(self.household_housing_attributes.columns)

        total_time = time.perf_counter() - start_time
        LOG.debug(
            f"Materialized alternatives into {human_shape(alternatives.shape)} array using {human_bytes(alternatives.nbytes)} in {human_time(total_time)}"
        )
        self.alternatives_colnames = (
            colnames  # hacky to set this every time but it never changes
        )

        if self.alternatives_stds is None:
            self.alternatives_stds = np.std(alternatives, axis=0)

        alternatives /= self.alternatives_stds

        return alternatives

    def create_alternatives(self):
        LOG.info("Creating alternatives")
        startTime = time.perf_counter()

        LOG.info("Converting pandas data to numpy")
        self.housing_xwalk = pd.Series(
            np.arange(len(self.housing_attributes)), index=self.housing_attributes.index
        )

        # we always have an unequilibrated choice to simplify coding, it is just only a single choice if not specified
        # good ol' mononomial logit model
        unequilibrated_choice = (
            self.unequilibrated_choice.copy()
            if self.unequilibrated_choice is not None
            else pd.Series(np.zeros(len(self.choice), index=self.choice.index))
        )
        unique_unequilibrated_choices = unequilibrated_choice.unique()
        self.unequilibrated_choice_xwalk = pd.Series(
            np.arange(len(unique_unequilibrated_choices)),
            index=unique_unequilibrated_choices,
        )
        self.hh_xwalk = pd.Series(
            np.arange(len(self.household_attributes)),
            index=self.household_attributes.index,
        )

        self.hh_hsg_choice = self.housing_xwalk.loc[
            self.choice.loc[self.hh_xwalk.index]
        ].values
        self.hh_unequilibrated_choice = self.unequilibrated_choice_xwalk.loc[
            self.unequilibrated_choice.loc[self.hh_xwalk.index]
        ].values

        if self.endogenous_variable_defs is not None:
            # neighborhood indices
            unique_neighborhoods = self.neighborhoods.unique()
            self.nbhd_xwalk = pd.Series(
                np.arange(len(unique_neighborhoods)), index=unique_neighborhoods
            )
            # neighborhood for each housing choice, neighborhood index for each neighborhood
            self.nbhd_for_choice = self.nbhd_xwalk.loc[
                self.neighborhoods.loc[self.housing_xwalk.index]
            ].values  # index is sorted
        else:
            self.nbhd_xwalk = self.nbhd_for_choice = None

        # index of each household in the full alternatives dataset
        # repeated for each alternative (housing choice)
        # so if there are three households, three housing choices, and three unequilibrated choices:
        # hhidx:         0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1 2 2 2 2 2 2 2 2 2
        # choiceidx:     0 0 0 1 1 1 2 2 2 0 0 0 1 1 1 2 2 2 0 0 0 1 1 1 2 2 2
        # uneqchoiceidx: 0 1 2 0 1 2 0 1 2 0 1 2 0 1 2 0 1 2 0 1 2 0 1 2 0 1 2
        # NB this could also be conceptualized as a four-dimensional array (household * housing choice * unequilibrated choice * variables)
        # but, uh, let's not
        LOG.info("Indexing full alternatives dataset")
        self.full_hhidx = np.repeat(
            np.arange(len(self.household_attributes)),
            len(self.housing_attributes) * len(unique_unequilibrated_choices),
        )
        self.full_choiceidx = np.repeat(
            np.tile(
                np.arange(len(self.housing_attributes)), len(self.household_attributes)
            ),
            len(unique_unequilibrated_choices),
        )
        self.full_uneqchoiceidx = np.tile(
            np.arange(len(unique_unequilibrated_choices)),
            len(self.housing_attributes) * len(self.household_attributes),
        )
        self.full_hsgchosen = self.hh_hsg_choice[self.full_hhidx] == self.full_choiceidx
        self.full_uneqchosen = (
            self.hh_unequilibrated_choice[self.full_hhidx] == self.full_uneqchoiceidx
        )
        self.full_chosen = self.full_hsgchosen & self.full_uneqchosen

        # create indices into the hh_housing_attributes so we can use it like a numpy array rather than with slow Pandas indexing
        # full_hh_hsg[i] is the 0-based index into household_housing_attributes for full_alternative i.
        # how we do it: build a Pandas series numbered 0...n and indexed like household_housing_attributes. then we zip the crosswalks together
        # and select from household_housing_attributes.
        if self.household_housing_attributes is not None:
            # do this in chunks to save memory
            self.full_hh_hsgidx = np.full_like(self.full_hhidx, -1, dtype="int32")
            hh_hsg_loc = pd.Series(
                np.arange(len(self.household_housing_attributes)),
                index=self.household_housing_attributes.index,
            )
            for chunk_start in range(0, len(self.full_hhidx), 100000):
                chunk_end = min(
                    chunk_start + 100000, len(self.household_housing_attributes)
                )
                self.full_hh_hsgidx[chunk_start:chunk_end] = hh_hsg_loc.loc[
                    list(
                        zip(
                            self.hh_xwalk.index[self.full_hhidx[chunk_start:chunk_end]],
                            self.housing_xwalk.index[
                                self.full_choiceidx[chunk_start:chunk_end]
                            ],
                        )
                    )
                ].values
            del hh_hsg_loc  # save memory
        else:
            self.full_hh_hsgidx = None

        # calculate endogenous variables if necessary (conditional is inside the function)
        self.initialize_or_update_endogenous_variables(initial=True)

        if self.sample_alternatives is None or self.sample_alternatives <= 0:
            if self.max_rent_to_income is None:
                self.alternatives = self.materialize_alternatives(
                    self.full_hhidx,
                    self.full_choiceidx,
                    self.full_uneqchoiceidx,
                    self.full_hh_hsgidx
                    if self.household_housing_attributes is not None
                    else None,
                )
                self.alternatives_hhidx = self.full_hhidx
                self.alternatives_choiceidx = self.full_choiceidx
                self.alternatives_uneqchoiceidx = self.full_uneqchoiceidx
                self.alternatives_hsgchosen = self.full_hsgchosen
                self.alternatives_uneqchosen = self.full_uneqchosen
                self.alternatives_chosen = self.full_chosen
            else:
                # TODO
                raise ValueError("max_rent_to_income with no sampling is unimplemented")

        else:
            LOG.info("Sampling alternatives")
            if self.max_rent_to_income is None:
                # note that we do not include the other unequilibrated choises for the chosen housing unit here, so they are not selected
                # randomly. We are randomly sampling housing alternatives, but always use all unequilibrated alternatives.
                feasible_unchosen_alts = ~self.full_hsgchosen
            else:
                feasible_unchosen_alts = (
                    self.income.astype("float64").values[self.full_hhidx]
                    * self.max_rent_to_income
                    > self.price.astype("float64").values[self.full_choiceidx]
                ) & ~self.full_hsgchosen

            # unequilibrated alternatives are not sampled
            n_housing_alts_per_hh = np.bincount(
                self.full_hhidx[feasible_unchosen_alts]
            ) / len(self.unequilibrated_choice_xwalk)

            def random_sel(n):
                if n <= self.sample_alternatives - 1:
                    return np.repeat([True], n)
                else:
                    ret = np.arange(n) < self.sample_alternatives - 1
                    self._rng.shuffle(ret)
                    return ret

            # since households are the outermost index, and unequilibrated choices are the innermost index, we can get away with this
            # the repeat() makes sure all unquilibrated alternatives are selected (recall they are always adjacent), and concatenating
            # is correct since households are the slowest-changing index
            sampled_mask = np.concatenate(
                [
                    np.repeat(random_sel(n), len(self.unequilibrated_choice_xwalk))
                    for n in n_housing_alts_per_hh
                ]
            )

            unchosen_sampled_idxs = np.arange(len(self.full_hhidx))[
                feasible_unchosen_alts
            ][sampled_mask]
            del sampled_mask
            chosen_idxs = np.arange(len(self.full_hhidx))[
                self.full_hsgchosen
            ]  # we do not sample uneq alternatives

            sampled_idxs = np.concatenate([unchosen_sampled_idxs, chosen_idxs])
            # put them back in the household > housing > uneq order
            np.sort(sampled_idxs)

            self.alternatives_hhidx = self.full_hhidx[sampled_idxs]
            self.alternatives_choiceidx = self.full_choiceidx[sampled_idxs]
            self.alternatives_uneqchoiceidx = self.full_uneqchoiceidx[sampled_idxs]
            self.alternatives_hsgchosen = self.full_hsgchosen[sampled_idxs]
            self.alternatives_uneqchosen = self.full_uneqchosen[sampled_idxs]
            self.alternatives_chosen = self.full_chosen[sampled_idxs]
            if self.household_housing_attributes is not None:
                self.alternatives_hh_hsgidx = self.full_hh_hsgidx[sampled_idxs]
            else:
                self.alternatives_hh_hsgidx = None
            self.alternatives = self.materialize_alternatives(
                self.alternatives_hhidx,
                self.alternatives_choiceidx,
                self.alternatives_uneqchoiceidx,
                self.alternatives_hh_hsgidx
                if self.household_housing_attributes is not None
                else None,
            )

        endTime = time.perf_counter()
        LOG.info(
            f"Created alternatives for {len(self.household_attributes)} households in {endTime - startTime:.3f} seconds"
        )
        LOG.info(f"Alternatives dimensions: {human_shape(self.alternatives.shape)}")
        LOG.info(f"Alternatives use {human_bytes(self.alternatives.nbytes)} memory")

    def full_utility(self, include_budget=True, include_ascs=True):
        """
        Calculate full utilities (i.e. utilities for all alternatives, not just sampled ones). To remain within available memory,
        this in chunks.
        """

        utility = np.full_like(self.full_choiceidx, np.nan, dtype="float64")
        # convert supply to an np array
        lnsupply = np.log(self.weighted_supply.loc[self.housing_xwalk.index].values)

        if self.price_income_transformation.n_params > 0:
            coefs = self.first_stage_fit.params.values[
                : -self.price_income_transformation.n_params
            ]
            price_income_params = self.first_stage_fit.params.values[
                -self.price_income_transformation.n_params :
            ]
        else:
            coefs = self.first_stage_fit.params.values
            price_income_params = np.zeros(0)

        chunk_rows = int(
            np.floor(self.max_chunk_bytes / len(self.alternatives_colnames) / 8)
        )  # bytes per float64
        nchunks = len(self.full_hhidx) // chunk_rows + 1
        LOG.info(
            f"Materializing full alternatives using {nchunks} chunks of {chunk_rows} rows each ({human_bytes(chunk_rows * len(self.alternatives_colnames) * 8)} each)"
        )

        nthreads = multiprocessing.cpu_count()

        task_queue = queue.Queue()
        result_queue = queue.Queue()
        # once set, all threads should stop
        stop_threads = threading.Event()

        # worker function to run in threads
        # numpy releases the GIL so running in threads works pretty well
        def worker():
            while not stop_threads.is_set():
                try:
                    # every 10 seconds restart the loop so we can check if thread has been signaled to stop
                    i, chunk_start, chunk_end = task_queue.get(timeout=10)
                except queue.Empty:
                    continue
                else:
                    chunk_alts = self.materialize_alternatives(
                        self.full_hhidx[chunk_start:chunk_end],
                        self.full_choiceidx[chunk_start:chunk_end],
                        self.full_uneqchoiceidx[chunk_start:chunk_end],
                        self.full_hh_hsgidx[chunk_start:chunk_end]
                        if self.household_housing_attributes is not None
                        else None,
                        price_income_params=price_income_params,
                    )

                    if not include_budget:
                        # zero out budget so it does not affect utility. chunk_alts has been materialized just for us, okay
                        # to be destructive.
                        chunk_alts[:, 0] = 0

                    # add systematic utility and deterministic part of ASC based on market share
                    # TODO okay to just add log(weighted) here when ASCs were calc'd with log(unweighted)? I think so.
                    util_chunk = (
                        np.dot(chunk_alts, coefs)
                        + lnsupply[self.full_choiceidx[chunk_start:chunk_end]]
                    )

                    result_queue.put((chunk_start, chunk_end, util_chunk))
                    if i % 10 == 9:
                        LOG.info(
                            f"Materialized chunk {i + 1} / {nchunks} and computed utility"
                        )
                    task_queue.task_done()

        LOG.info(f"computing utility using {nthreads} threads")

        # Thread to put everything into the utility function
        def consumer():
            while not stop_threads.is_set():
                try:
                    # every 10 seconds restart the loop so we can check if thread has been signaled to stop
                    chunk_start, chunk_end, util_chunk = result_queue.get(timeout=10)
                except queue.Empty:
                    continue
                else:
                    utility[chunk_start:chunk_end] = util_chunk
                    result_queue.task_done()

        threads = [
            threading.Thread(target=worker, daemon=False) for i in range(nthreads)
        ]
        threads.append(threading.Thread(target=consumer, daemon=False))
        for thread in threads:
            thread.start()

        for i, chunk_start in enumerate(range(0, len(self.full_hhidx), chunk_rows)):
            chunk_end = min(chunk_start + chunk_rows, len(self.full_hhidx))
            task_queue.put((i, chunk_start, chunk_end))

        # wait for everything to finish
        task_queue.join()
        result_queue.join()

        # signal all threads to stop
        stop_threads.set()

        if include_ascs:
            utility += (
                self.first_stage_ascs.values[self.full_choiceidx]
                + self.first_stage_uneq_ascs.values[self.full_uneqchoiceidx]
            )

        return utility

    def fit_first_stage(self):
        LOG.info("fitting first stage")

        self._log_supply = np.log(np.bincount(self.hh_hsg_choice))[
            self.alternatives_choiceidx
        ]

        self.first_stage_fit = MNLFullASC(
            alternatives=self.alternatives,
            choiceidx=(self.alternatives_choiceidx, self.alternatives_uneqchoiceidx),
            hhidx=self.alternatives_hhidx,
            chosen=self.alternatives_chosen,
            supply=(
                np.bincount(self.hh_hsg_choice),
                np.bincount(self.hh_unequilibrated_choice),
            ),
            starting_values=np.zeros(self.alternatives.shape[1], dtype="float64"),
            param_names=[
                *self.alternatives_colnames,
                *self.price_income_transformation.param_names,
            ],
            method=self.method,
            minimize_options=self.minimize_options,
            est_ses=self.est_first_stage_ses,
            add_ln_supply=(0,),
        )

        self.first_stage_fit.fit()

        # recalculate ASCs to clear full market
        # to make the model tractable, we estimate on a sample of the alternatives - which provides consistent but inefficient parameter estimates
        # (see Ben-Akiva and Lerman 1985)
        # Since when we clear the market in scenario evaluation, we use the full set of alternatives, not just the sampled alternatives, we want
        # the baseline (current conditions) to be market clearing as well, we re-estimate the ASCs with the full set of alternatives
        # another way to look at this is that due to tractability concerns we cannot estimate the full model without alternative sampling,
        # so we lose some efficiency - but we can estimate the ASCs without alternative sampling, so we don't lose the efficiency there.

        if self.max_rent_to_income is None:
            feasible_alts = np.full(True, len(self.full_hhidx))
        else:
            feasible_alts = (
                self.income.astype("float64").values[self.full_hhidx]
                * self.max_rent_to_income
                > self.price.astype("float64").values[self.full_choiceidx]
            )

        base_utility = self.full_utility(include_ascs=False)

        assert not np.any(np.isnan(base_utility[feasible_alts]))

        fullAscStartTime = time.perf_counter()

        ascs = compute_ascs(
            base_utilities=base_utility[feasible_alts],
            supply=(
                # of homes
                self.weighted_supply.loc[self.housing_xwalk.index].values,
                # of unequilibrated choices
                # While we don't (obviously) equilibrate unequilibrated choices, we do use their supply to find their ASCs in the first-stage fit
                np.bincount(
                    self.hh_unequilibrated_choice,
                    self.weights.loc[self.hh_xwalk.index].values,
                ),
            ),
            hhidx=self.full_hhidx[feasible_alts],
            choiceidx=(
                self.full_choiceidx[feasible_alts],
                self.full_uneqchoiceidx[feasible_alts],
            ),
            starting_values=self.first_stage_fit.ascs,
            weights=self.weights.loc[self.hh_xwalk.index].values,
            log=True,
        )
        fullAscEndTime = time.perf_counter()

        # descale coefs
        # but don't descale transformation parameters
        # recall that the _result_ of the transformation, not the inputs, is what is scaled, so this is okay
        # TODO make sure this is rescaling correctly
        scalars = pd.Series(self.alternatives_stds, index=self.alternatives_colnames)
        scalars.loc[self.price_income_transformation.param_names] = 1
        scalars = scalars.reindex(self.first_stage_fit.params.index)
        self.first_stage_fit.params /= scalars
        if self.est_first_stage_ses and self.first_stage_fit.se is not None:
            self.first_stage_fit.se /= scalars

        # no scaling needed anymore
        self.alternatives_stds = np.ones_like(self.alternatives_stds)

        self.first_stage_ascs = pd.Series(ascs[0], index=self.housing_xwalk.index)
        self.first_stage_uneq_ascs = pd.Series(
            ascs[1], index=self.unequilibrated_choice_xwalk.index
        )
        LOG.info(
            f"Finding full ASCs took {human_time(fullAscEndTime - fullAscStartTime)}"
        )

        # don't serialize
        self.first_stage_fit.alternatives = None
        self.first_stage_fit._log_supply = None

    def fit_second_stage(self):
        LOG.info("fitting second stage")
        startTime = time.perf_counter()
        second_stage_exog = sm.add_constant(
            self.housing_attributes[self.second_stage_params]
        )
        second_stage_endog = self.first_stage_ascs.reindex(second_stage_exog.index)

        mod = sm.OLS(second_stage_endog, second_stage_exog)
        self.second_stage_fit = mod.fit()
        self.type_shock = self.second_stage_fit.resid
        endTime = time.perf_counter()
        LOG.info(f"Fit second stage in {endTime - startTime:.2f} seconds")

    def fit(self):
        self.fit_first_stage()
        if self.second_stage_params is not None:
            self.fit_second_stage()
        else:
            LOG.info("No second stage requested")

    def sort(self, maxiter=np.inf):
        """
        Clear the market with a change to supply

        TODO document which data structures can be changed and still have this function return correct results
        """
        LOG.info("Clearing the market and sorting households")

        # convert supply to an np array
        supply = self.weighted_supply.loc[self.housing_xwalk.index].values

        # allow for slight floating point error several orders of magnitude smaller than convergence criterion
        if np.abs(np.sum(supply) - np.sum(self.weights)) > 1e-8:
            raise ValueError(
                f"total supply has changed! expected {np.sum(self.weights)} but found {np.sum(supply)}"
            )

        # first update second stage
        if self.second_stage_params is not None:
            LOG.info("updating second stage")
            pred_ascs = (
                self.second_stage_fit.predict(
                    sm.add_constant(self.housing_attributes[self.second_stage_params])
                )
                + self.type_shock
            )

            maxabsdiff = np.max(np.abs(pred_ascs - self.first_stage_ascs))
            LOG.info(
                f"Second stage updated with changes to first-stage ASCs of up to {maxabsdiff:.2f}"
            )
            self.first_stage_ascs = pred_ascs
        else:
            LOG.info("No second stage fit, not updating")

        clear_market = ClearMarket(self)
        converged = clear_market.clear_market()
        if not converged:
            raise ConvergenceError("Prices did not converge!")

    def initialize_or_update_endogenous_variables(self, initial: bool) -> None:
        """
        Initialize endogenous variables, or update them based on predicted probabilities after the model has been fit.
        :param initial: if True, will compute endogenous variables based on observed choices. If False, will compute
            based on predicted probabilities
        :type initial: bool
        """
        if self.endogenous_variable_defs is None:
            LOG.info("not updating endogenous variables as none are defined")
        else:
            if initial:
                # initial fit: use actual chosen values to compute endogenous variables
                LOG.info("creating endogenous variables")
                weighted_probs = self.full_hsgchosen
            else:
                # if model is already fitted, update endogenous variables with fitted values
                LOG.info("updating endogenous variables")
                weighted_probs = self._probabilities()

            if self.weights is not None:
                # if weights are present, use them
                weighted_probs = (
                    weighted_probs
                    * self.weights.loc[self.hh_xwalk.index].values[self.full_hhidx]
                )

            # if this is the initial run, initialize endogenous_varnames
            # need to materialize and save because dict.keys() order is not guaranteed, though in recent versions of
            # Python it is stable (in fact, may have always been stable?)
            if initial:
                self.endogenous_varnames = list(self.endogenous_variable_defs.keys())

            # Neighborhood for every element of weighted_probs
            nbhdidx = self.nbhd_for_choice[self.full_choiceidx]

            # reset endogenous variables. this creates a new array. if we instead cleared the old array, we would need
            # to create a copy in the if block below to ensure that the original values were preserved for comparison
            # purposes
            self.endogenous_variables = np.full(
                (np.max(self.nbhd_for_choice) + 1, len(self.endogenous_varnames)),
                np.nan,
                "float64",
            )

            # compute endogenous variables
            for neighborhood in range(np.max(self.nbhd_for_choice) + 1):
                nbhdmask = nbhdidx == neighborhood
                # all probabilities for the neighborhood
                nbhdweights = weighted_probs[nbhdmask]
                # how likely every household is to choose this neighborhood
                hhweights = np.bincount(self.full_hhidx[nbhdmask], weights=nbhdweights)
                # use the precomputed list to avoid order changing, as iteration order in python dicts is not guaranteed
                # endogenous_varnames is created by initialize_endogenous_variables
                for i, varname in enumerate(self.endogenous_varnames):
                    func = self.endogenous_variable_defs[varname]
                    self.endogenous_variables[neighborhood, i] = func(
                        self.household_attributes, self.income, hhweights
                    )

            # check our work
            assert not np.any(
                np.isnan(self.endogenous_variables)
            ), "some endogenous variables are nan!"

            if initial:
                # save the initial variables for later comparison
                # this array is never modified, an entirely new array is created when sorting, so no need to copy
                self.orig_endogenous_variables = self.endogenous_variables

    def _probabilities(self):
        "Compute probabilities and return as numpy array, use .probabilities() for a Pandas data frame"
        LOG.info("finding utility")

        if self.max_rent_to_income is None:
            feasible_alts = np.full(True, len(self.full_hhidx))
        else:
            feasible_alts = (
                self.income.astype("float64").values[self.full_hhidx]
                * self.max_rent_to_income
                > self.price.astype("float64").values[self.full_choiceidx]
            )

        exp_utility = np.exp(self.full_utility()[feasible_alts])
        assert not np.any(np.isnan(exp_utility))
        expsums = np.bincount(self.full_hhidx[feasible_alts], exp_utility)
        probs = np.zeros_like(self.full_choiceidx, dtype="float64")
        # alts not in choice set b/c infeasible are left with probability zero
        probs[feasible_alts] = exp_utility / expsums[self.full_hhidx[feasible_alts]]
        return probs

    def probabilities(self):
        "Return choice probabilities as Pandas dataframe, with columns for household ID, choice ID, and uneq choice ID"
        probs = self._probabilities()
        return pd.DataFrame(
            {
                "hh": self.hh_xwalk.index[self.full_hhidx],
                "housing": self.housing_xwalk.index[self.full_choiceidx],
                "uneq_choice": self.unequilibrated_choice_xwalk.index[
                    self.full_uneqchoiceidx
                ],
                "probability": probs,
            }
        )

    def _mkt_shares(self):
        probs = (
            self._probabilities()
            * self.weights.loc[self.hh_xwalk.index].values[self.full_hhidx]
        )
        return np.bincount(self.full_choiceidx, probs)

    def mkt_shares(self):
        return pd.Series(self._mkt_shares(), index=self.housing_xwalk.index)

    def _uneq_mkt_shares(self):
        probs = (
            self._probabilities()
            * self.weights.loc[self.hh_xwalk.index].values[self.full_hhidx]
        )
        return np.bincount(self.full_uneqchoiceidx, probs)

    def uneq_mkt_shares(self):
        return pd.Series(
            self._uneq_mkt_shares(), index=self.unequilibrated_choice_xwalk.index
        )

    def savenew(self, basefile):
        save_load.save(basefile, self)

    @classmethod
    def loadnew(cls, basefile):
        return save_load.load(basefile)

    def to_text(self, fn=None):
        "Save model results as text. If fn==None, return as string"

        outstring = """Equilibrium sorting model (Tra [2007] formulation, price in first stage as budget constraint)
Model run initiated at {creation_time}
Budget function {price_income_transformation}

First stage (discrete choice sorting model):
{first_stage_summary}

Equilibrated (housing) ASCs:
{equilibrated_ascs}

Unequilibrated ASCs:
{unequilibrated_ascs}

Second stage (OLS parameters):
{second_stage_summary}

Fit with EqSorMo version {version}, https://github.com/mattwigway/eqsormo
        """.format(
            first_stage_summary=self.first_stage_fit.summary(),
            unequilibrated_ascs=pd.DataFrame(self.first_stage_uneq_ascs).to_string(),
            equilibrated_ascs=pd.DataFrame(
                self.first_stage_ascs.describe()
            ).to_string(),
            second_stage_summary=self.second_stage_fit.summary(),
            creation_time=self.creation_time.strftime("%Y-%m-%d %H:%M:%S %Z"),
            price_income_transformation=self.price_income_transformation.name,
            version=eqsormo.version,
        )

        if fn is not None:
            with open(fn, "w") as outfile:
                outfile.write(outstring)
        else:
            return outstring

    # don't pickle fullAlternatives
    def __getstate__(self):
        return {k: v for k, v in self.__dict__.items() if k != "fullAlternatives"}

    @classmethod
    def from_pickle(cls, fn):
        tra = super().from_pickle(fn)
        tra.create_full_alternatives()
        return tra



#    Copyright 2020 Matthew Wigginton Conway

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# Author: Matthew Wigginton Conway <matt@indicatrix.org>
#         School of Geographical Sciences and Urban Planning
#         Arizona State University

# Lazy-load arrays from an NPZ file instead of loading everything at once

import numpy as np
from zipfile import ZipFile
import tempfile
import os


class LazyNPZ(object):
    """
    Lazy load Numpy arrays from an NPZ file.
    """

    def __init__(self, file_or_filename, allow_pickle=False):
        self.zipfile = ZipFile(file_or_filename)
        self.tempdir = tempfile.mkdtemp(prefix="npz")
        self.allow_pickle = allow_pickle

    def get_members(self):
        # strip off .npy
        return [i[:-4] for i in self.zipfile.namelist()]

    def get_member(self, item, mmap=False):
        "get a member, possibly loading it using mmap to save memory"
        if mmap:
            return self._get_member_mmap(item)
        else:
            return self._get_member(item)

    def _get_member(self, item):
        "get a member and return as bona-fide array"
        with self.zipfile.open(item + ".npy") as f:
            return np.load(f, allow_pickle=self.allow_pickle)

    def _get_member_mmap(self, item):
        "get a member and return as an mmapped-array"
        # extract the npy file
        npypath = self.zipfile.extract(item + ".npy", path=self.tempdir)
        return np.load(npypath, mmap_mode="r+", allow_pickle=self.allow_pickle)

    def __del__(self):
        self.zipfile.close()
        # can't delete temporary directory here, may be in use by mmapped objects
        # TODO figure out how to delete temp directory when all mmapped arrays are closed



#    Copyright 2019-2020 Matthew Wigginton Conway

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# Author: Matthew Wigginton Conway <matt@indicatrix.org>, School of Geographical Sciences and Urban Planning, Arizona State University

# TODO Do we need this file anymore?

from logging import getLogger
import time
import statsmodels.api as sm
import pandas as pd
import numpy as np
import scipy.optimize, scipy.stats
import linearmodels
from tqdm import tqdm
import pickle
from .ascs import compute_ascs

LOG = getLogger(__name__)


class SortingModel(object):
    """
    Represents a random utility based equilibrium sorting model.

    Based on the model described in Klaiber and Phaneuf (2010) "Valuing open space in a residential sorting model of the Twin Cities"
    https://doi.org/10.1016/j.jeem.2010.05.002 (no open access version available, unfortunately)
    """

    def __init__(
        self,
        altCharacteristics,
        altPrice,
        hh,
        hhChoice,
        interactions,
        initialPriceCoef,
        sampleAlternatives=None,
        method="bfgs",
    ):
        """
        :param altCharacteristics: Home and neighborhood attributes of housing alternatives/types (assumed that all alternatives are available to all households)
        :type altHousing: Pandas dataframe
        
        :param altPrice: Equilibrium (non-instrumented) price of housing alternatives (will be replaced with instrumented version)
        :type altPrice: Pandas series of float64, indexed like altHousing and altNeighborhood

        :param hh: Attributes of households
        :type hh: Pandas dataframe

        :param hhChoice: Alternative actually chosen by households
        :type hhChoice: Pandas series, indexed like hh, with values matching index of altHousing et al.

        :param interactions: Which sociodemographics should be interacted with which housing or neighborhood attributes.
        :type interactions: iterable of tuple (sociodemographic column, housing/neighborhood column)

        :param initialPriceCoef: For the 2SLS estimate, an initial estimate for the price coefficient is needed; this value is iterated from and should converge.
        :type intialPriceCoef: float

        :param sampleAlternatives: If a positive integer, take a random sample of this size for available alternatives to estimate the model. Recommended in models with a large number of alternatives. If 0 or None, no sampling will be performed. Use np.set_seed() before calling fit for model reproducibility when using this option.
        :type sampleAlternatives: int or None

        :param method: A scipy.optimize.minimize method used for maximizing log-likelihood of first stage, default 'bfgs'
        """

        # Protective copies of all the things
        self.altCharacteristics = altCharacteristics.astype("float64").copy()
        self.price = altPrice.copy()
        self.hh = hh.copy().apply(
            lambda col: col - col.mean()
        )  # predemean household characteristics
        self.hh.index.name = "household"  # for joining convenience later
        self.hhChoice = hhChoice.copy()
        self.interactions = interactions
        self.initialPriceCoef = initialPriceCoef
        self.sampleAlternatives = sampleAlternatives

        #: supply by housing type
        self.supply = self.hhChoice.value_counts().astype("float64")

        self.validate()

        assert not "price" in self.altCharacteristics.columns
        self.method = method

    def validate(self):
        "Check for obvious errors in model inputs"
        LOG.warn(
            "Validation function not implemented. You better be darn sure of your inputs."
        )

    def fit(self):
        "Fit the whole model"
        startTime = time.perf_counter()
        LOG.info("Fitting equilibrium sorting model")
        if self.alternatives is None:
            raise RuntimeError("Alternatives not yet created")

        self.fit_first_stage()
        self.fit_second_stage()
        endTime = time.perf_counter()
        LOG.info(
            f"""
Fitting sorting model took {endTime - startTime:.3f} seconds.
Convergence:
  First stage: {self.first_stage_converged}
"""
        )

    def create_alternatives(self):
        LOG.info("Creating alternatives")
        startTime = time.perf_counter()

        self.fullAlternatives = pd.concat(
            [self.altCharacteristics for i in range(len(self.hh))], keys=self.hh.index
        )
        self.fullAlternatives["chosen"] = False
        self.fullAlternatives["hhchoice"] = self.hhChoice.reindex(
            self.fullAlternatives.index, level=0
        )
        self.fullAlternatives.loc[
            self.fullAlternatives.index.get_level_values(1)
            == self.fullAlternatives.hhchoice,
            "chosen",
        ] = True

        LOG.info("created full set of alternatives, now sampling if requested")

        if self.sampleAlternatives <= 0 or self.sampleAlternatives is None:
            self.alternatives = self.fullAlternatives
        else:
            unchosenAlternatives = (
                self.fullAlternatives[~self.fullAlternatives.chosen]
                .groupby(level=0)
                .apply(lambda x: x.sample(self.sampleAlternatives - 1))
            )
            unchosenAlternatives.index = unchosenAlternatives.index.droplevel(
                0
            )  # fix dup'd household level due to groupby
            self.alternatives = pd.concat(
                [
                    unchosenAlternatives,
                    self.fullAlternatives[self.fullAlternatives.chosen],
                ]
            ).sort_index(level=[0, 1])

        self.alternatives.drop(columns=["chosen"], inplace=True)
        self.fullAlternatives.drop(columns=["chosen"], inplace=True)

        endTime = time.perf_counter()
        LOG.info(
            f"Created {len(self.alternatives)} alternatives for {len(self.hh)} in {endTime - startTime:.3f} seconds"
        )

    def first_stage_utility(self, params, mean_indirect_utility):
        # TODO I don't think that adding the mean_indirect_utility this way will work from an indexing standpoint
        # diffs is differences due to sociodemographics from mean indirect utility
        diffs = self.firstStageData.multiply(params, axis="columns").sum(axis="columns")
        utilities = (
            diffs
            + mean_indirect_utility.loc[
                self.firstStageData.index.get_level_values("choice")
            ].values
        )
        return utilities

    def first_stage_probabilities(self, params, mean_indirect_utility):
        utility = self.first_stage_utility(params, mean_indirect_utility)
        expUtility = np.exp(utility)
        if not np.all(np.isfinite(expUtility)):
            raise ValueError(
                f"Household/choice combinations {expUtility.index[~np.isfinite(expUtility)]} have non-finite utilities!"
            )
        return expUtility / expUtility.groupby(level=0).sum()

    def compute_mean_indirect_utility(self, params):
        # These are the alternative specific constants, which are not fit by ML, but rather using a contraction mapping that lets
        # the model converge faster - see Equation 16 of Bayer et al. (2004).
        mean_indirect_utility = compute_ascs(
            self.firstStageData, params, self.supply, self._prev_mean_indirect_utility
        )
        self._prev_mean_indirect_utility = mean_indirect_utility
        return mean_indirect_utility

    def first_stage_neg_loglikelihood(self, params):
        mean_indirect_utility = self.compute_mean_indirect_utility(params)
        logprobs = np.log(self.first_stage_probabilities(params, mean_indirect_utility))
        return -np.sum(logprobs.loc[list(zip(self.hhChoice.index, self.hhChoice))])

    def fit_first_stage(self):
        "Perform the first stage estimation"
        LOG.info("Performing first-stage estimation")

        startTime = time.perf_counter()

        self.firstStageData = pd.DataFrame()

        # demean sociodemographics so ASCs are interpretable as mean indirect utility (or something like that... TODO check)
        altsWithHhCharacteristics = self.alternatives.join(
            self.hh
        )  # should project to all alternatives TODO check

        for interaction in self.interactions:
            self.firstStageData[f"{interaction[0]}_{interaction[1]}"] = (
                altsWithHhCharacteristics[interaction[0]]
                * altsWithHhCharacteristics[interaction[1]]
            )

        # solve scaling issues
        stdevs = self.firstStageData.apply(np.std)
        self.firstStageData = self.firstStageData.divide(stdevs, axis="columns")

        LOG.info(f"Fitting {len(self.firstStageData.columns)} interaction parameters")

        self._prev_mean_indirect_utility = pd.Series(
            np.zeros(len(self.altCharacteristics)), index=self.altCharacteristics.index
        )

        self.first_stage_loglik_constants = -self.first_stage_neg_loglikelihood(
            np.zeros(len(self.firstStageData.columns))
        )

        minResults = scipy.optimize.minimize(
            self.first_stage_neg_loglikelihood,
            np.zeros(len(self.firstStageData.columns)),
            method=self.method,
            options={"disp": True},
        )

        self.first_stage_loglik_beta = -self.first_stage_neg_loglikelihood(minResults.x)
        self.interaction_params = (
            pd.Series(minResults.x, self.firstStageData.columns) / stdevs
        )  # correct the scaling
        self.interaction_params_se = (
            pd.Series(
                np.sqrt(np.diag(minResults.hess_inv)), self.firstStageData.columns
            )
            / stdevs
        )
        # TODO robust SEs
        self.mean_indirect_utility = self.compute_mean_indirect_utility(
            self.interaction_params
        )
        self.first_stage_converged = minResults.success

        endTime = time.perf_counter()
        if self.first_stage_converged:
            LOG.info(
                f"First stage converged in {endTime - startTime:.3f} seconds: {minResults.message}"
            )
        else:
            LOG.error(
                f"First stage FAILED TO CONVERGE in {endTime - startTime:.3f} seconds: {minResults.message}"
            )

    def fit_second_stage(self):
        "Fit the instrumental variables portion of the model"
        LOG.info("Fitting second stage")

        startTime = time.perf_counter()

        priceCoef = prevPriceCoef = self.initialPriceCoef

        iter = 0
        with tqdm() as pbar:
            while True:
                residual_utility = self.mean_indirect_utility - priceCoef * self.price
                # Constant should be included since location of ASCs is arbitrary, see Klaiber and Kuminoff (2014) note 9
                olsreg = sm.OLS(
                    residual_utility, sm.add_constant(self.altCharacteristics)
                )
                olsfit = olsreg.fit()

                # compute the price instrument by solving for prices that clear the market - which means the prices that produce the mean indirect utilities found
                # in the first stage, since those are the market-clearing utilities.
                priceIv = (self.mean_indirect_utility - olsfit.fittedvalues) / priceCoef

                ivreg = linearmodels.IV2SLS(
                    self.mean_indirect_utility,
                    sm.add_constant(self.altCharacteristics),
                    pd.DataFrame(self.price.rename("price")),
                    pd.DataFrame(priceIv.rename("price_iv")),
                )
                self.second_stage_fit = ivreg.fit()

                priceCoef = self.second_stage_fit.params.price

                iter += 1
                pbar.update()
                if np.abs(priceCoef - prevPriceCoef) < 1e-6:
                    endTime = time.perf_counter()
                    LOG.info(
                        f"Price coefficient converged in {endTime-startTime:.3f} seconds after {iter} iterations"
                    )
                    self.mean_params = self.second_stage_fit.params
                    self.mean_params_se = (
                        self.second_stage_fit.std_errors
                    )  # NB these are wrong b/c they don't account for variation in theta
                    self.type_shock = (
                        self.second_stage_fit.resids
                    )  # TODO ensure this does not include residuals from the first stage 2SLS estimates
                    self.price_iv = priceIv
                    break
                else:
                    prevPriceCoef = priceCoef

    def summary(self):
        # summarize params
        summary = pd.DataFrame(
            {
                "coef": pd.concat([self.interaction_params, self.mean_params]),
                "se": pd.concat(
                    [self.interaction_params_se, self.mean_params_se]
                ),  # TODO no standard errors for second stage yet...
            }
        )

        summary["z"] = summary.coef / summary.se
        # TODO should probably be t-test, but then I have to do a bunch of df calculations...
        summary["p"] = (1 - scipy.stats.norm.cdf(np.abs(summary.z))) * 2
        return summary

    def rebuild_full_alternatives_with_interactions(self):
        "Call after attributes of alternatives have changed to get utilities right"
        self.fullAlternativesWithInteractions = self.fullAlternatives.join(self.hh)

        for left, right in self.interactions:
            self.fullAlternativesWithInteractions[f"{left}_{right}"] = (
                self.fullAlternativesWithInteractions[left]
                * self.fullAlternativesWithInteractions[right]
            )

        self.fullTypeShock = self.type_shock.reindex(
            self.fullAlternativesWithInteractions.index, level=1
        )
        self.fullAlternativesWithInteractions["price"] = self.price.reindex(
            self.fullAlternativesWithInteractions.index, level=1
        )

    def utilities(self):
        "Get utilities of every household choosing every house type"
        # compute utilities, without unobserved price shocks
        params = pd.concat([self.mean_params, self.interaction_params])
        utilities = (
            sm.add_constant(self.fullAlternativesWithInteractions)[params.index]
            .multiply(params, "columns")
            .sum(axis="columns")
        )
        utilities += self.fullTypeShock

        # solve numerical problems by making all utilities positive
        # you can add a constant to all utilities for a particular household and get the same probabilities
        # utilities -= utilities.groupby(level=0).min().reindex(utilities.index, level=0)

        return utilities

    # work in percentages rather than probabilities to avoid underflow issues
    def probabilities(self, utilities=None):
        if utilities is None:
            utilities = self.utilities()

        expUtilities = np.exp(utilities)
        if not np.all(np.isfinite(expUtilities)):
            LOG.warn("some utilities are not finite when exponentiated")
        if not np.all(expUtilities > 0):
            LOG.warn("some utilities are zero when exponentiated")
        logsums = expUtilities.groupby(level=0).sum()  # group by households and sum
        percentages = (expUtilities) / logsums.reindex(expUtilities.index, level=0)
        return percentages

    def market_shares(self, utilities=None):
        return self.probabilities(utilities=utilities).groupby(level=1).sum()

    def clear_market(self):
        "Adjust prices so the market clears"
        prices = self.price  # for comparison later

        LOG.info("Clearing the market (everyone stand back)")

        # utilities with constants - pre compute for performance
        mean_indirect_utility = compute_ascs(
            self.fullAlternativesWithInteractions,
            self.interaction_params,
            self.supply,
            self.mean_indirect_utility,
        )

        # solve for price
        nonPriceParams = self.mean_params.drop(["price"])
        nonPriceObsUtilities = (
            sm.add_constant(self.altCharacteristics)[nonPriceParams.index]
            .multiply(nonPriceParams, axis="columns")
            .sum(axis="columns")
        )
        self.price = (
            mean_indirect_utility - nonPriceObsUtilities - self.type_shock
        ) / self.mean_params.price
        self.fullAlternativesWithInteractions["price"] = self.price.reindex(
            self.fullAlternativesWithInteractions.index, level=1
        )

        maxPriceChange = np.max(np.abs(self.price - prices))
        LOG.info(
            f"Market cleared with absolute price changes of up to {maxPriceChange}"
        )
        LOG.info(f"New price distribution: {self.price.describe()}")

    def to_pickle(self, fn):
        "Save a fit model to disk"
        if isinstance(fn, str):
            with open(fn, "wb") as out:
                pickle.dump(self, out)
        else:
            pickle.dump(self, fn)

    @staticmethod
    def from_pickle(fn):
        "Read a previously saved model"
        if isinstance(fn, str):
            with open(fn, "rb") as inf:
                model = pickle.load(inf)
        else:
            model = pickle.load(fn)

        if not isinstance(model, SortingModel):
            raise ValueError("File does not contain a sorting model!")

        return model



#    Copyright 2019-2020 Matthew Wigginton Conway

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# Author: Matthew Wigginton Conway <matt@indicatrix.org>, School of Geographical Sciences and Urban Planning, Arizona State University

import numpy as np
from logging import getLogger

LOG = getLogger(__name__)


# Numba does not speed this up
# TODO convergence_criterion doesn't do anything
def compute_ascs(
    base_utilities,
    supply,
    hhidx,
    choiceidx,
    starting_values=None,
    convergence_criterion=1e-6,
    weights=None,
    log=False,
):
    """
    Compute the alternative specific constants (ASCs) that should be added to base_utilities to make the market shares equal supply.

    This allows multinomial logit models with a full set of ASCs to converge a lot faster. For models with full sets of ASCs,
    the maximum-likelihood estimate will reproduce market shares perfectly. This means that the ASCs are implied by any set of coefficients.
    This algorithm finds those implied ASCs based on the "base" utilities provided - which are the utilities calculated based on the parameters,
    but without ASCs.

    This is based on equation 16 of Bayer et al (2004)

    :param base_utilities: Vector of utilities calculated using the current parameters and variable values, but no ASCs
    :type base_utilities: numpy.ndarray

    :param supply: supply of each alternative along each margin
    :type supply: list of numpy.ndarray

    :param hhidx: household index associated with each utility
    :type hhidx: numpy.ndarray

    :param choiceidx: housing type index associated with each utility
    :type choiceidx: numpy.ndarray

    :param starting_values: starting values for ASCs, default all zeros
    :type starting_values: list with numpy.ndarray with length equal to number of choices on each margin

    :param weights: household weights, indexed by hhidx (NOTE THAT SUPPLY SHOULD BE WEIGHTED AS WELL WHEN USING WEIGHTS)
    :type weights: numpy.ndarray
    """

    if starting_values is None:
        ascs = [np.zeros(chcidx.max() + 1) for chcidx in choiceidx]
    else:
        ascs = starting_values

    # this variable is unhelpfully named to increase efficiency, so that we do not have to allocate as many (potentially
    # large) arrays. at various points in the algorithm it contains utilities, then exponentiated utilities, then
    # probabilities
    values = np.copy(base_utilities)
    while True:
        values[:] = base_utilities
        # values now contains base utilities
        for margin in range(len(ascs)):
            np.add(values, ascs[margin][choiceidx[margin]], out=values)
        # values now contains full utilities with ASCs

        np.exp(values, out=values)
        # values now contains exponentiated utilities
        if np.any(~np.isfinite(values)):
            # TODO should raise ValueError, but that breaks numba
            print(
                "Some exponentiated utilities are non-finite! This may be a scaling issue."
            )
            # print('Max ASC:')
            # print(np.max(ascs))
            # print('Min ASC')
            # print(np.min(ascs))
            # print('NaNs:')
            # print(np.sum(np.isnan(ascs)))
            # print('Min utility')
            # print(np.min(first_stage_utilities))
            # print('Max utility')
            # print(np.max(first_stage_utilities))
            # print('Min exp(utility)')
            # print(np.min(exp_utils))
            # print('Max exp(utility)')
            # print(np.max(exp_utils))
            return [
                np.array([42.0])
            ]  # will cause errors somewhere else, so the process will crash, and hopefully the user
            # will find the output of the above print statement while debugging.

        logsums = np.bincount(hhidx, weights=values)
        np.divide(values, logsums[hhidx], out=values)
        # values now contains probabilities

        if weights is not None:
            # multiply each prob by the weight of the household it represents
            np.multiply(values, weights[hhidx], out=values)
        # values now contains weighted probabilities

        converged = True
        first_stage_shares = []

        # check convergence on all dimensions
        for margin in range(len(ascs)):
            margin_shares = np.bincount(choiceidx[margin], weights=values)
            first_stage_shares.append(margin_shares)

            if np.abs(np.sum(margin_shares) - np.sum(supply[margin])) > 1e-3:
                # TODO should raise ValueError, but that breaks numba
                print(
                    "Total demand does not equal total supply! This may be a scaling issue."
                )
                return [
                    np.array([42.0])
                ]  # will cause errors somewhere else, so the process will crash, and hopefully the user
                # will find the output of the above print statement while debugging.

            if not np.allclose(margin_shares, supply[margin]):
                converged = False
                if log:
                    LOG.info(
                        f"Margin {margin}: maxdiff: {np.max(margin_shares - supply[margin])}, mindiff: {np.min(margin_shares - supply[margin])} {len(ascs[margin])} ascs"
                    )

        if converged:
            break
        else:
            # update ASCs
            # this is not done in the above for loop because once the model has converged we don't want to update any more ASCs
            for margin in range(len(ascs)):
                ascs[margin] = ascs[margin] - np.log(
                    first_stage_shares[margin] / supply[margin]
                )

                # normalize, can add/subtract constant to utility and not change predictions
                # this is true even in the multidimensional case, because _every_ outcome changes simultaneously
                ascs[margin] -= ascs[margin][0]

    return ascs



#    Copyright 2019-2020 Matthew Wigginton Conway

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# Author: Matthew Wigginton Conway <matt@indicatrix.org>, School of Geographical Sciences and Urban Planning, Arizona State University

import numpy as np
import scipy.optimize
import scipy.stats
import time
from logging import getLogger
import pandas as pd
import threading
import multiprocessing
import queue

from .compute_ascs import compute_ascs
from .util import human_time

LOG = getLogger(__name__)


class MNLFullASC(object):
    """
    A multinomial logit model with full ASCs. With full ASCs, the model reproduces market shares perfectly. This
    can be exploited so that the maximum likelihood algorithm does not have to actually find the ASCs - for any set
    of coefficients, the ASCs that reproduce market shares perfectly are implied. This speeds estimation significantly.

    This class estimates such multinomial logit models.
    """

    def __init__(
        self,
        alternatives,
        supply,
        hhidx,
        choiceidx,
        chosen,
        starting_values,
        method="L-BFGS-B",
        asc_convergence_criterion=1e-6,
        param_names=None,
        est_ses=True,
        minimize_options={},
        add_ln_supply=(),
    ):
        """
        :param utility: Function that returns utility for all choice alternatives and decisionmakers for a given set of parameters, not including ASCs. Receives parameters as a numpy array. Should return a MultiIndexed-pandas dataframe with indices called 'decisionmaker' and 'choice'
        :type utility: function

        :param supply: List of supply of each choice in equilibrium, for each choice margin (choice margins may be simple or joint)
        :type supply: List of np.array

        :param choiceidx: List of arrays of what choice a particular utility is for, on each margin
        :type choiceidx: List of np.array

        :param starting_values: starting values for parameters
        :type starting_values: numpy array

        :param method: scipy optimize method, default 'bfgs'

        :param minimize_options: dict of additional options for selected minimization method
        :type minimize_options: dict

        :param add_ln_supply: margins (parallel to supply) that should have log of supply added to utility
        :type add_ln_supply: tuple
        """
        self.alternatives = (
            alternatives  # TODO will this always be called with self as first argument?
        )
        self.supply = supply
        self.starting_values = starting_values
        self.method = method
        self.asc_convergence_criterion = asc_convergence_criterion
        self.param_names = param_names
        self._previous_ascs = None
        self.asc_time = 0
        self.hhidx = hhidx
        self.choiceidx = choiceidx
        self.chosen = chosen
        self.est_ses = est_ses
        self.minimize_options = minimize_options
        self.add_ln_supply = add_ln_supply

        # precompute the ln(supply) values to be added to utility
        self._log_supply = np.zeros_like(choiceidx[0], dtype="float64")
        for dim in add_ln_supply:
            self._log_supply += np.log(supply[dim][choiceidx[dim]])

    def base_utility(self, params):
        "Utility without ASCs"
        return np.dot(self.alternatives, params).reshape(-1) + self._log_supply

    def compute_ascs(self, base_utilities):
        """
        Compute alternative specific constants implied by base_utilities.

        Uses a contraction mapping found in equation 16 of Bayer et al 2004.
        """
        startTime = time.perf_counter()

        ascs = compute_ascs(
            base_utilities,
            self.supply,
            self.hhidx,
            self.choiceidx,
            starting_values=self._previous_ascs,
            convergence_criterion=self.asc_convergence_criterion,
        )
        self._previous_ascs = ascs  # speed convergence later

        endTime = time.perf_counter()
        totalTime = endTime - startTime
        # LOG.info(f'found ASCs in {human_time(totalTime)}')
        self.asc_time += totalTime
        return ascs

    def full_utility(self, params):
        "Full utilities including ASCs"
        base_utilities = self.base_utility(params)
        # LOG.info(f'Computed base utilities in {human_time(end_time - start_time)}')
        ascs = self.compute_ascs(base_utilities)
        full_utilities = base_utilities
        for margin in range(len(ascs)):
            # okay to use += here, base utilities will not be used again
            full_utilities += ascs[margin][self.choiceidx[margin]]
        return full_utilities

    def probabilities(self, utility):
        exp_utility = np.exp(utility)
        if not np.all(np.isfinite(exp_utility)):
            raise ValueError(
                "Household/choice combinations "
                + str(exp_utility.index[~np.isfinite(exp_utility)])
                + " have non-finite utilities!"
            )
        logsums = np.bincount(self.hhidx, weights=exp_utility)
        return exp_utility / logsums[self.hhidx]

    def choice_probabilities(self, utility):
        probabilities = self.probabilities(utility)
        return probabilities[self.chosen]

    def negative_log_likelihood_for_utility(self, utility):
        negll = -np.sum(np.log(self.choice_probabilities(utility)))
        if np.isnan(negll) or not np.isfinite(negll):
            LOG.warn(
                "log-likelihood nan or not finite"
            )  # just warn, solver may be able to get out of this
        # LOG.info(f'Current negative log likelihood: {negll:.2f}')
        return negll

    def negative_log_likelihood(self, params):
        return self.negative_log_likelihood_for_utility(self.full_utility(params))

    # some optimizers call with a second state arg, some do not. be lenient.
    def log_progress(self, params, state=None):
        negll = self.negative_log_likelihood(params)
        improvement = negll - self._prev_ll
        LOG.info(
            f"After iteration {self._iteration}, -ll {negll:.7f}, change: {improvement:.7f}"
        )
        self.negll_for_iteration.append(negll)
        self.params_for_iteration.append(np.copy(params))
        self._prev_ll = negll
        self._iteration += 1

    def fit(self):
        LOG.info("Fitting multinomial logit model")
        startTime = time.perf_counter()

        # this is in fact the log likelihood at constants, because all ASCs are still estimated
        # Note that this is only true if starting values are all zeros
        if not np.allclose(self.starting_values, 0):
            LOG.warn(
                "not all starting values are zero, log likelihood at constants is actually log likelihood at starting values and may be incorrect"
            )
        self.loglik_constants = -self.negative_log_likelihood(self.starting_values)

        self._iteration = 0
        self._prev_ll = -self.loglik_constants
        self.params_for_iteration = []
        self.negll_for_iteration = []
        LOG.info(f"Before iteration 0, -ll {self._prev_ll:.7f}")
        minResults = scipy.optimize.minimize(
            self.negative_log_likelihood,
            self.starting_values,
            method=self.method,
            options={"disp": True, **self.minimize_options},
            callback=self.log_progress,
        )
        self.params_for_iteration = np.array(self.params_for_iteration)
        self.negll_for_iteration = np.array(self.negll_for_iteration)

        if self.est_ses:
            if minResults.success:
                try:
                    LOG.info("calculating Hessian")
                    # TODO robust SEs
                    # hess = statsmodels.tools.numdiff.approx_hess3(minResults.x, self.negative_log_likelihood)
                    # hess = approx_hess3(minResults.x, self.negative_log_likelihood)
                    hess = self.linear_hessian(minResults.x)
                    LOG.info("inverting Hessian")
                    hessInv = np.linalg.inv(hess)
                    ses = np.sqrt(np.diag(hessInv))
                    LOG.info("done calculating and inverting Hessian")
                except KeyboardInterrupt:
                    self.est_ses = False
                    LOG.warn(
                        "Keyboard interrupt caught, not calculating standard errors"
                    )
            else:
                LOG.error("Not calculating Hessian because model did not converge")
                self.est_ses = False
        else:
            LOG.info(
                "Not calculating Hessian because standard errors were not requested"
            )

        self.loglik_beta = -self.negative_log_likelihood(minResults.x)
        if self.param_names is None:
            self.params = minResults.x
            if self.est_ses:
                self.se = ses
            else:
                self.se = None
        else:
            self.params = pd.Series(minResults.x, index=self.param_names)
            if self.est_ses:
                self.se = pd.Series(ses, index=self.param_names)
            else:
                self.se = None

        # TODO compute t-stats
        if self.est_ses:
            self.zvalues = self.params / self.se
            # two-tailed test
            self.pvalues = 2 * scipy.stats.norm.cdf(-np.abs(self.zvalues))

        self.ascs = self.compute_ascs(self.base_utility(minResults.x))
        self.converged = minResults.success

        endTime = time.perf_counter()
        if self.converged:
            LOG.info(
                f"Multinomial logit model converged in {human_time(endTime - startTime)}: {minResults.message}"
            )
        else:
            LOG.error(
                f"Multinomial logit model FAILED TO CONVERGE in {human_time(endTime - startTime)}: {minResults.message}"
            )
        LOG.info(f"  Finding ASCs took {human_time(self.asc_time)}")

    def linear_hessian(self, params, step_size=None):
        """
        Compute the Hessian, quickly. This takes advantage of the fact that many components of utility are constant
        when computing the Hessian, which changes two parameters at a time. The ASCs also do not really need to be
        recomputed, as the changes to the coefficients are miniscule. This only works for utility functions that are
        strictly linear-in-parameters, which the MNL utility function is.

        The results of this function should be identical to statsmodels.tools.numdiff.approx_hess3, as the algorithm
        used is the same (equation 9 from Ridout 2009 - note that Ridout 2009 is about complex step differentiation, but
        eq. 9 is an example of finite-step differentiation).

        Ridout, M. S. (2009). Statistical Applications of the Complex-Step Method of Numerical Differentiation.
            _The American Statistician_, 63(1), 66–74. https://doi.org/10.1198/tast.2009.0013
        """
        utility = self.full_utility(params)

        # this is the step size for the finite-difference approximation. it is copied from
        # statsmodels.tools.numdiff.approx_hess3
        if step_size is None:
            step_size = np.MachAr().eps ** (1.0 / 4) * np.maximum(np.abs(params), 0.1)
        elif isinstance(step_size, float):
            step_size = np.full(len(params), step_size)

        nthreads = multiprocessing.cpu_count()
        LOG.info(f"computing Hessian using {nthreads} threads")
        # we need the outer product to scale the finite diff appx below
        prod = np.outer(step_size, step_size)
        hess = np.zeros_like(prod)

        task_queue = queue.Queue()
        result_queue = queue.Queue()
        stop_threads = threading.Event()

        n = len(params)

        # parallelized by rows, each thread calcs rows at a time, b/c there is some computation that's common across
        # a row.
        def worker():
            while not stop_threads.is_set():
                try:
                    i = task_queue.get(timeout=10)
                except queue.Empty:
                    continue
                else:
                    step_i = self.alternatives[:, i] * step_size[i]
                    LOG.info(f"Hessian starting row {i} / {n}")
                    for j in range(i, n):
                        step_j = self.alternatives[:, j] * step_size[j]
                        elem = (
                            self.negative_log_likelihood_for_utility(
                                utility + step_i + step_j
                            )
                            - self.negative_log_likelihood_for_utility(
                                utility + step_i - step_j
                            )
                            - (
                                self.negative_log_likelihood_for_utility(
                                    utility - step_i + step_j
                                )
                                - self.negative_log_likelihood_for_utility(
                                    utility - step_i - step_j
                                )
                            )
                        ) / (4 * prod[i, j])

                        result_queue.put((i, j, elem))
                    # only mark as done once per row
                    task_queue.task_done()

        def consumer():
            while not stop_threads.is_set():
                try:
                    i, j, elem = result_queue.get(timeout=10)
                except queue.Empty:
                    continue
                else:
                    hess[i, j] = hess[j, i] = elem
                    result_queue.task_done()

        for i in range(nthreads):
            threading.Thread(target=worker).start()
        threading.Thread(target=consumer).start()

        for i in range(n):
            task_queue.put(i)

        task_queue.join()
        result_queue.join()
        stop_threads.set()

        return hess

    def summary(self):
        with pd.option_context("display.max_rows", None):
            notes = []
            if self.est_ses:
                summ = pd.DataFrame(
                    {
                        "coef": self.params,
                        "se": self.se,
                        "z": self.zvalues,
                        "p": self.pvalues,
                    }
                ).round(3)
            else:
                summ = pd.DataFrame({"coef": self.params}).round(3)
                notes.append("Standard errors not estimated")

            if not self.converged:
                notes.append("WARNING: CONVERGENCE NOT ACHIEVED.")

            chi2 = 2 * (self.loglik_beta - self.loglik_constants)
            chi2df = len(self.params)
            pchi2 = 1 - scipy.stats.chi2.cdf(chi2, df=chi2df)

            # get around f-string cannot include backslash error
            newline = "\n"

            return f"""
Multinomial logit model with full ASCs
Parameters:
{str(summ)}
Notes:
{newline.join(notes)}

Log likelihood at constants: {self.loglik_constants:.3f}
Log likelihood at convergence: {self.loglik_beta:.3f}
Pseudo R-squared (McFadden): {1 - self.loglik_beta / self.loglik_constants:.3f}
Chi2 (vs constants): {chi2}, {chi2df} degrees of freedom
P(Chi2): {pchi2}
            """

        # don't pickle big arrays
        def __getstate__(self):
            return {
                k: v
                for k, v in self.__dict__.items()
                if k not in {"hhidx", "choiceidx", "chosen"}
            }



#    Copyright 2020 Matthew Wigginton Conway

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

byte_scalars = {"TB": 1e12, "GB": 1e9, "MB": 1e6, "KB": 1e3}


def human_bytes(nbytes):
    "Return bytes in human-readable format, e.g. 1.42GB"
    for suffix, scalar in byte_scalars.items():
        if nbytes > scalar:
            return f"{nbytes / scalar:.2f} {suffix}"

    return f"{nbytes} bytes"


def human_time(seconds):
    "Time in human readable format, e.g. 3630 -> 1h0m30s"
    ret = []
    if seconds >= 3600:
        ret.append(f"{int(seconds // 3600)}h")
        seconds = seconds % 3600
    if seconds >= 60:
        ret.append(f"{int(seconds // 60)}m")
        seconds = seconds % 60
    ret.append(f"{seconds:.3f}s")

    return " ".join(ret)


def human_shape(shape):
    "Shape of a numpy array in human-readable format, e.g. 2x2"
    return "x".join(map(str, shape))


class ConvergenceError(Exception):
    pass



#    Copyright 2019 Matthew Wigginton Conway

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# Author: Matthew Wigginton Conway <matt@indicatrix.org>, School of Geographical Sciences and Urban Planning, Arizona State University

# TODO is this file still needed?

import numpy as np
from tqdm import tqdm
import pandas as pd


def compute_ascs(data, params, supply, starting_values, convergence_criterion=1e-6):
    """
    Compute the ASCs implied by params for a multinomial logit model using data so that supply is correct.
    Data should be indexed by decisionmakers, then by choices.

    The algorithm herein is defined in Bayer et al. (2004), in equation 16. Specifically, for any set of parameters, the set
    starting values for the ASCs can be moved closer to the true values using the formula

    ASC_h(t+1) = ASC_h(t) - ln(sum(P_h) - S_h) where P_h is the probability of a particular decisionmaker (subscript suppressed) choosing outcome h.

    Repeat this process until the supply and demand are within convergence_criterion for every choice
    """

    # Utilities without the ASCs
    if isinstance(params, pd.Series):
        baseUtilities = (
            data[params.index].multiply(params, axis="columns").sum(axis="columns")
        )
    else:  # assume numpy array
        baseUtilities = data.multiply(params, axis="columns").sum(axis="columns")

    ascs = starting_values

    while True:
        firstStageUtilities = (
            baseUtilities
            + ascs.loc[baseUtilities.index.get_level_values("choice")].values
        )
        expUtils = np.exp(firstStageUtilities)
        firstStageShares = (
            (expUtils / expUtils.groupby(level=0).sum()).groupby(level=1).sum()
        )
        if np.abs(firstStageShares.sum() - supply.sum()) > 1e-3:
            raise ValueError(
                "Total demand does not equal total supply! This may be a scaling issue."
            )
        if np.max(np.abs(firstStageShares - supply)) < convergence_criterion:
            break
        ascs = ascs - np.log(firstStageShares / supply)

        # normalize, can add/subtract constant to utility and not change predictions
        ascs -= ascs.iloc[0]

    return ascs



#    Copyright 2020 Matthew Wigginton Conway

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# Author: Matthew Wigginton Conway <matt@indicatrix.org>
#         School of Geographical Sciences and Urban Planning,
#         Arizona State University

import numpy as np
import dill
import logging
import uuid
from eqsormo.common.lazy_npz import LazyNPZ

MODEL_VERSION = 0

LOG = logging.getLogger(__name__)

# This is very hacky, but what we are doing here is saving two files, on with numpy arrays in it, and one with
# everything else. This will hopefully work until we can come up with something better...

FIELDS = [
    "alternatives_colnames",
    "alternatives_stds",
    "choice",
    "creation_time",
    "endogenous_variable_defs",
    "endogenous_variables",
    "endogenous_varnames",
    "est_first_stage_ses",
    "first_stage_ascs",
    "first_stage_fit",
    "first_stage_uneq_ascs",
    "fixed_price",
    "full_choiceidx",
    "full_chosen",
    "full_hh_hsgidx",
    "full_hhidx",
    "full_hsgchosen",
    "full_uneqchoiceidx",
    "full_uneqchosen",
    "hh_hsg_choice",
    "hh_unequilibrated_choice",
    "hh_xwalk",
    "household_attributes",
    "household_housing_attributes",
    "housing_attributes",
    "housing_xwalk",
    "income",
    "interactions",
    "max_chunk_bytes",
    "max_rent_to_income",
    "method",
    "minimize_options",
    "nbhd_xwalk",
    "nbhd_for_choice",
    "neighborhoods",
    "orig_endogenous_variables",
    "orig_price",
    "price",
    "price_income_starting_values",
    "price_income_transformation",
    "sample_alternatives",
    "second_stage_fit",
    "second_stage_params",
    "seed",
    "type_shock",
    "unequilibrated_choice",
    "unequilibrated_choice_xwalk",
    "unequilibrated_hh_params",
    "unequilibrated_hsg_params",
    "weighted_supply",
    "weights",
]

# fields that should be mmapped to save memory
# these are large fields that we only use part of
MMAP_FIELDS = [
    "full_choiceidx",
    "full_chosen",
    "full_hh_hsgidx",
    "full_hhidx",
    "full_hsgchosen",
    "full_uneqchoiceidx",
    "full_uneqchosen",
    "hh_hsg_choice",
    "hh_unequilibrated_choice",
    "hh_xwalk",
    "household_housing_attributes",
]


def save(basefile, model):
    model_uuid = uuid.uuid4().hex

    pickle_fields = {"MODEL_VERSION": MODEL_VERSION, "MODEL_ID": model_uuid}
    numpy_fields = {
        "MODEL_VERSION": np.array([MODEL_VERSION]),
        "MODEL_ID": np.array([model_uuid]),
    }

    for field in FIELDS:
        if hasattr(model, field):
            val = getattr(model, field)
            if isinstance(val, np.ndarray):
                numpy_fields[field] = val
            else:
                pickle_fields[field] = val
        else:
            LOG.warn(f"Field {field} not found in model to be serialized")

    with open(f"{basefile}.pickle", "wb") as out:
        dill.dump(pickle_fields, out)

    # TODO any way to set allow_pickle=False here?
    np.savez_compressed(f"{basefile}.npz", **numpy_fields)


def load(basefile):
    # do here to prevent circular import
    from .tra_sorting_model import TraSortingModel

    with open(f"{basefile}.pickle", "rb") as inp:
        pkl = dill.load(inp)

    npz = LazyNPZ(f"{basefile}.npz", allow_pickle=False)
    npz_contents = npz.get_members()

    if (
        pkl["MODEL_VERSION"] != MODEL_VERSION
        or npz.get_member("MODEL_VERSION")[0] != MODEL_VERSION
    ):
        # TODO npz model version not saved correctly so not checked
        raise ValueError(
            f'Model version {pkl["MODEL_VERSION"]} does not match model version {MODEL_VERSION}'
        )

    if pkl["MODEL_ID"] != npz.get_member("MODEL_ID")[0]:
        raise ValueError("Model UUIDs do not match")

    def get(field):
        if field in pkl:
            return pkl[field]
        elif field in npz_contents:
            return npz.get_member(field, mmap=field in MMAP_FIELDS)
        else:
            LOG.warn(f"Field {field} not found in stored model, assuming it was none")
            return None

    model = TraSortingModel(
        housing_attributes=get("housing_attributes"),
        household_attributes=get("household_attributes"),
        interactions=get("interactions"),
        unequilibrated_hh_params=get("unequilibrated_hh_params"),
        unequilibrated_hsg_params=get("unequilibrated_hsg_params"),
        second_stage_params=get("second_stage_params"),
        price=get("price"),
        income=get("income"),
        choice=get("choice"),
        unequilibrated_choice=get("unequilibrated_choice"),
        price_income_transformation=get("price_income_transformation"),
        price_income_starting_values=get("price_income_starting_values"),
        sample_alternatives=get("sample_alternatives"),
        method=get("method"),
        max_rent_to_income=get("max_rent_to_income"),
        household_housing_attributes=get("household_housing_attributes"),
        weights=get("weights"),
        minimize_options=get("minimize_options"),
        neighborhoods=get("neighborhoods"),
        endogenous_variable_defs=get("endogenous_variable_defs"),
        max_chunk_bytes=get("max_chunk_bytes"),
        est_first_stage_ses=get("est_first_stage_ses"),
        seed=get("seed"),
        fixed_price=get("fixed_price"),
    )

    for field in FIELDS:
        setattr(model, field, get(field))

    return model



#    Copyright 2019 Matthew Wigginton Conway

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# Author: Matthew Wigginton Conway <matt@indicatrix.org>, School of Geographical Sciences and Urban Planning, Arizona State University

"""
Functional forms for price and income
"""

import numpy as np
import scipy.stats


class FunctionalForm(object):
    def __init__(self, func, name, n_params=0, starting_values=None, param_names=None):
        """
        Create a new functional form. Pass in a function of (self, income, price, *args) that returns a transformed
        income/price variable. *args is the parameters for the function, and will have length n_params, which will start at
        the values in starting_values, which are then estimated in the maximum likelihood estimation. Note that income and price
        are numpy arrays or Pandas series, not scalars.
        """
        self.apply = func
        self.n_params = n_params
        self.name = name

        if param_names is None:
            param_names = [f"price_income_param_{i}" for i in range(n_params)]
        self.param_names = param_names

        if starting_values is None:
            starting_values = np.zeros(n_params)
        self.starting_values = starting_values

        assert len(param_names) == n_params
        assert len(starting_values) == n_params

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


# Log difference, i.e. ln(income - price), as used in Tra (2010)
# TODO: price > income? or drop from choice set
logdiff = FunctionalForm(
    lambda income, price: np.log(income - price), "log(income - price)"
)

# normcdf, kinda like log but defined below zero
# assumes ppl are most price sensitive where budget = price
normdiff = FunctionalForm(
    lambda income, price, scale: scipy.stats.norm.cdf(income - price, scale=scale),
    "normcdf(income - price)",
    n_params=1,
    starting_values=np.array([1]),
    param_names=["normcdf_sd"],
)

# square root of difference, as used in Tra (2010) robustness checks
sqrtdiff = FunctionalForm(
    lambda income, price: np.sqrt(income - price), "sqrt(income - price)"
)

# box-cox transform of income - price
# Box and Cox (1964)
# NB any price > income should be taken care of by a max_rent_to_income parameter
# TODO can we remove the max rent to income param here?
boxcoxdiff = FunctionalForm(
    lambda income, price, lmbda: scipy.stats.boxcox(
        np.maximum(income - price, 1), lmbda=lmbda
    ),
    "boxcox(income - price)",
    n_params=1,
    param_names=["boxcox_lambda"],
)

# Form suggested by Stephane Hess in his advanced discrete choice modelling class, London, UK, July 2019
# Note that this will not behave right if called on a subset of the data, since it uses the mean of income
# Removed because these functions are now always called on subsets
# hess = FunctionalForm(lambda income, price, hess_lambda: ((income / np.mean(income)) ** hess_lambda) * price,
#     'Hess: (income / mean(income)) ^ lambda * price', n_params=1, param_names=['hess_lambda'], starting_values=[-1])



"""
Classes to implement the variety of sorting model described in Tra CI (2013) Measuring the General Equilibrium Benefits of Air Quality Regulation in Small Urban Areas. Land Economics 89(2): 291–307.
"""

from .tra_sorting_model import TraSortingModel
from . import price_income



#    Copyright 2019-2020 Matthew Wigginton Conway

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import logging

# from .sorting_model import SortingModel
from .tra import TraSortingModel
from . import tra

version = "0.8.0"

rootLogger = None


def enable_logging():
    """
    Enable logging for the EqSorMo package.

    EqSorMo uses the Python logging package. If you are managing loggers at a higher level in a large project, you may not want to run this
    function. Otherwise, running this function will create a root logger and enable logging to the console.
    """
    global rootLogger

    if rootLogger is None:
        rootLogger = logging.getLogger()
        rootLogger.setLevel(logging.INFO)
        rootLogger.addHandler(logging.StreamHandler())

        rootLogger.info(f"Eqsormo version {version}")
    else:
        rootLogger.info("Logging already enabled.")



#    Copyright 2019 Matthew Wigginton Conway

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# Author: Matthew Wigginton Conway <matt@indicatrix.org>, School of Geographical Sciences and Urban Planning, Arizona State University

import dill
import pickle


class BaseSortingModel(object):
    """
    Represents a random utility based equilibrium sorting model.

    Based on the model described in Klaiber and Phaneuf (2010) "Valuing open space in a residential sorting model of the Twin Cities"
    https://doi.org/10.1016/j.jeem.2010.05.002 (no open access version available, unfortunately)
    """

    def to_pickle(self, fn):
        "Save a fit model to disk"
        if isinstance(fn, str):
            with open(fn, "wb") as out:
                dill.dump(self, out)
        else:
            dill.dump(self, fn)

    @classmethod
    def from_pickle(cls, fn):
        "Read a previously saved model"
        if isinstance(fn, str):
            with open(fn, "rb") as inf:
                model = dill.load(inf)
        else:
            model = dill.load(fn)

        if not isinstance(model, cls):
            raise ValueError("File does not contain a sorting model!")

        return model



#    Copyright 2019-2021 Matthew Wigginton Conway

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# Author: Matthew Wigginton Conway <matt@indicatrix.org>
#         School of Geographical Sciences and Urban Planning,
#         Arizona State University

import numpy as np

from logging import getLogger
import tempfile
import queue
import threading
import pandas as pd
import os
import time
import scipy.optimize
import multiprocessing
from functools import lru_cache
from eqsormo.common.util import human_time

LOG = getLogger(__name__)


class ClearMarket(object):
    def __init__(self, model, price_step=1e-5, maxiter=None):
        self.model = model
        self.price = model.price.loc[model.housing_xwalk.index].to_numpy()
        self.supply = model.weighted_supply.loc[model.housing_xwalk.index].to_numpy()
        self.fixed_price_index = model.housing_xwalk.loc[model.fixed_price]
        self.fixed_price = model.price.loc[model.fixed_price]
        self.price_step = price_step
        self.maxiter = maxiter
        self.alt_income = model.income.loc[model.hh_xwalk.index].to_numpy()[
            model.full_hhidx
        ]

    def clear_market(self, diagonal_iterations=3):
        """
        Parameters
        ----------
        diagonal_iterations: int, optional
            Number of iterations to perform with the diagonal of the Jacobian before switching to the full Jacobian.
            The diagonal of the Jacobian is much faster to compute, but does not guarantee convergence, but generally moves in
            the right direction. Doing the first few interations with the diagonal can significantly speed convergence.
        """
        LOG.info("Clearing the market (everyone stand back)")
        start_time = time.perf_counter()

        if self.model.endogenous_variable_defs is not None:
            raise ValueError("Endogeneous variables not supported in sorting")

        self.non_price_utilities = self.model.full_utility(include_budget=False)

        i = 0
        current_price = self.remove_fixed_price(self.price)

        LOG.info("Computing shares")
        shares = self.shares(current_price)
        alpha = 1
        while self.maxiter is None or i < self.maxiter:
            # make the logs more consistent
            i += 1
            LOG.info(f"market clearing: begin iteration {i}")
            excess_demand = shares - self.supply
            current_obj_val = np.sum(excess_demand ** 2)

            # since they always have to sum to 100% of hhs max will always be >= 0, and min <= 0
            LOG.info(
                f"Maximum overdemand: {np.max(excess_demand):.3f}, underdemand: {np.min(excess_demand):.3f}, SSE: {current_obj_val}"
            )
            LOG.info(
                f"Maximum overdemand: {np.max(excess_demand / self.supply) * 100:.3f}%, underdemand: {np.min(excess_demand / self.supply) * 100:.3f}%"
            )

            # update prices
            jacob = self.compute_derivatives(
                current_price, shares, diagonal_only=i <= diagonal_iterations
            )
            LOG.info("Inverting Jacobian")
            jacob_inv = np.linalg.inv(jacob)

            # this is 7.7 from Tra's dissertation
            price_delta_full = jacob_inv @ self.remove_fixed_price(excess_demand)

            # run a golden section-ish search to minimize excess demand as much as possible using this set of price deltas
            LOG.info(
                f"Previous alpha value: {alpha}. Using Brent's method to find optimal alpha given gradient."
            )

            # hacky, but use an LRU cache here to avoid recomputing shares after optimization converges
            @lru_cache(maxsize=10)
            def prices_and_shares_for_alpha(candidate_alpha):
                candidate_prices = current_price - price_delta_full * candidate_alpha
                candidate_shares = self.shares(candidate_prices)
                return candidate_prices, candidate_shares

            def obj_val(candidate_alpha):
                new_shares = prices_and_shares_for_alpha(candidate_alpha)[1]
                return np.sum((new_shares - self.supply) ** 2)

            # Do not run many iterations here, don't spend too much time doing this
            alpha_res = scipy.optimize.minimize_scalar(
                obj_val,
                bracket=(alpha, alpha * 0.9),
                tol=1e-3,
                options={"disp": True, "maxiter": 5, "xtol": 1e-3},
                method="brent",
            )
            alpha = (
                alpha_res.x
            )  # don't actually care if it converged, as long as it moved us closer

            new_price, new_shares = prices_and_shares_for_alpha(alpha)
            new_obj_val = np.sum((new_shares - self.supply) ** 2)

            LOG.info(f"Found optimal alpha {alpha}, moves SSE from {current_obj_val} to {new_obj_val}")

            if not new_obj_val < current_obj_val:
                if i <= diagonal_iterations:
                    LOG.error(
                        f"With optimal alpha {alpha}, objective did not improve, using full Jacobian!"
                    )
                    diagonal_iterations = -1
                    continue
                else:
                    raise ValueError(f"Objective did not improve with alpha {alpha}!")

            current_price = new_price
            shares = new_shares

            if np.allclose(shares, self.supply):
                self.model.price = self.to_pandas_price(
                    self.add_fixed_price(current_price)
                )
                end_time = time.perf_counter()
                LOG.info(
                    f"Market clearing converged in {i} iterations after {human_time(end_time - start_time)}"
                )
                return True

        # can only get here if maxiter is reached
        else:
            end_time = time.perf_counter()
            LOG.info(
                f"Market clearing FAILED TO CONVERGE in {self.maxiter} iterations after {human_time(end_time - start_time)}."
            )
            return False

    def shares(self, price):
        budgets, feasible_alts = self.get_budgets(price)
        full_utilities = (
            self.non_price_utilities
            + self.model.first_stage_fit.params.budget * budgets
        )
        exp_utility = np.exp(full_utilities)
        del full_utilities, budgets
        # force choice probability to zero for infeasible alts
        exp_utility[~feasible_alts] = 0
        del feasible_alts

        if not np.all(np.isfinite(exp_utility)):
            raise FloatingPointError("Not all exp(utilities) are finite (scaling?)")

        expsums = np.bincount(self.model.full_hhidx, weights=exp_utility)
        probs = exp_utility / expsums[self.model.full_hhidx]

        if self.model.weights is not None:
            probs *= self.model.weights.loc[self.model.hh_xwalk.index].to_numpy()[
                self.model.full_hhidx
            ]

        shares = np.bincount(self.model.full_choiceidx, weights=probs)

        return shares

    def get_budgets(self, price):
        """
        Return the budgets as well as the feasible alternatives for a set of prices.
        """
        alt_price = self.add_fixed_price(price)[self.model.full_choiceidx]
        budget = np.zeros_like(self.alt_income)
        if self.model.max_rent_to_income is None:
            feasible_alts = np.full_like(alt_price, True)
        else:
            feasible_alts = self.alt_income * self.model.max_rent_to_income > alt_price

        budget = np.zeros_like(self.alt_income)
        budget[feasible_alts] = self.model.price_income_transformation.apply(
            self.alt_income[feasible_alts],
            alt_price[feasible_alts],  # TODO price income params
        )

        return budget, feasible_alts

    def remove_fixed_price(self, price):
        """
        From a full price vector, return a price vector with the fixed price removed, used in market clearing.
        Since one price is held constant, we don't feed it into the root-finding algorithm.
        """
        return price[
            np.r_[0 : self.fixed_price_index, self.fixed_price_index + 1 : len(price)]
        ]

    def add_fixed_price(self, price):
        """
        From a price vector resulting from root finding, add the fixed price back in
        """
        # insert the fixed price at location fixed_price_index
        return np.concatenate(
            (
                price[: self.fixed_price_index],
                [self.fixed_price],
                price[self.fixed_price_index :],
            )
        )

    def to_pandas_price(self, price):
        """
        Convert a price vector _which includes the fixed price_ back to Pandas format.
        """
        return pd.Series(price, index=self.model.housing_xwalk.index)

    def compute_derivatives(self, price, base_shares, diagonal_only=False):
        if diagonal_only:
            LOG.info("Ignoring off-diagonal elements of Jacobian")

        jacob = np.zeros((len(price), len(price)))

        budgets, feasible_alts = self.get_budgets(price)
        budget_coef = self.model.first_stage_fit.params.budget
        base_exp_utilities = np.exp(self.non_price_utilities + budget_coef * budgets)
        # exp utility of zero is out of choice set
        base_exp_utilities[~feasible_alts] = 0
        del budgets, feasible_alts

        if not np.all(np.isfinite(base_exp_utilities)):
            raise FloatingPointError("Not all exp(utilities) are finite (scaling?)")

        budget_step, feasible_alts_step = self.get_budgets(price + self.price_step)
        budget_coef = self.model.first_stage_fit.params.budget
        step_exp_utilities = np.exp(
            self.non_price_utilities + budget_coef * budget_step
        )
        # exp utility of zero is out of choice set
        step_exp_utilities[~feasible_alts_step] = 0
        del budget_step, feasible_alts_step

        if not np.all(np.isfinite(step_exp_utilities)):
            raise FloatingPointError(
                "Not all exp(step_utilities) are finite (scaling?)"
            )

        # memmap the base utilities so we can use copy-on-write
        fh, util_file = tempfile.mkstemp(prefix="eqsormo_mmap_", suffix=".bin")
        os.close(fh)
        try:
            mm = np.memmap(
                util_file, dtype="float64", mode="w+", shape=base_exp_utilities.shape
            )
            mm[:] = base_exp_utilities[:]
            mm.flush()

            # cache weights
            if self.model.weights is not None:
                alt_weights = self.model.weights.loc[
                    self.model.hh_xwalk.index
                ].to_numpy()[self.model.full_hhidx]

            task_queue = queue.Queue()
            result_queue = queue.Queue()
            stop_threads = threading.Event()

            def worker():
                while not stop_threads.is_set():
                    try:
                        jacidx, choice = task_queue.get(timeout=10)
                    except queue.Empty:
                        continue  # return to top of loop to check for stop_threads event
                    else:
                        exp_utilities = np.memmap(
                            util_file,
                            dtype="float64",
                            mode="c",
                            shape=step_exp_utilities.shape,
                        )
                        # save memory by just saving indices, not full bool array
                        # in the model in my dissertation, there are ~1000 choices, so this will be 1/1000 of the indices
                        # since an int is 32 bits (or maybe 64 since numpy arrays can be big), and a bool I believe takes up
                        # an entire byte, we come out ahead b/c 1/1000 * 32 < 8
                        choicemask = np.nonzero(self.model.full_choiceidx == choice)
                        exp_utilities[choicemask] = step_exp_utilities[choicemask]
                        util_sums = np.bincount(
                            self.model.full_hhidx, weights=exp_utilities
                        )

                        if diagonal_only:
                            probs = (
                                exp_utilities[choicemask]
                                / util_sums[self.model.full_hhidx[choicemask]]
                            )
                            if self.model.weights is not None:
                                probs *= alt_weights[choicemask]
                            share_step = np.sum(probs)
                            jaccol = np.zeros(len(price))
                            jaccol[jacidx] = (
                                share_step - base_shares[choice]
                            ) / self.price_step
                        else:
                            probs = exp_utilities / util_sums[self.model.full_hhidx]
                            if self.model.weights is not None:
                                probs *= alt_weights
                            share_step = np.bincount(
                                self.model.full_choiceidx, weights=probs
                            )
                            del probs
                            jaccol = self.remove_fixed_price(
                                (share_step - base_shares) / self.price_step
                            )

                        result_queue.put((jacidx, jaccol))
                        task_queue.task_done()

            def consumer():
                while not stop_threads.is_set():
                    try:
                        jacidx, jaccol = result_queue.get(timeout=10)
                    except queue.Empty:
                        continue
                    else:
                        if jacidx % 10 == 9:
                            LOG.info(f"computed derivative {jacidx + 1} / {len(price)}")
                        jacob[:, jacidx] = jaccol
                        result_queue.task_done()

            # might need something more complex here to account for the memory pressure of sorting. Even if you have
            # 16 cores, you might not have enough memory to compute 16 derivatives at once.
            nthreads = multiprocessing.cpu_count() if diagonal_only else 2
            LOG.info(f"computing derivatives using {nthreads} threads")

            # start threads
            for i in range(nthreads):
                threading.Thread(target=worker, daemon=False).start()

            # consumer thread
            threading.Thread(target=consumer, daemon=False).start()

            # fill queue
            # note that jacidx is the index in the Jacobian, which is not the same as choice which is the housing choice
            # b/c one housing choice is skipped in the Jacobian
            for jacidx, choice in enumerate(
                np.r_[
                    0 : self.fixed_price_index,
                    self.fixed_price_index + 1 : len(self.price),
                ]
            ):
                task_queue.put((jacidx, choice))

            # await completion
            task_queue.join()
            result_queue.join()

            assert not np.any(
                np.diag(jacob) >= 0
            ), "some diagonal elements of jacobian are nonnegative!"

            # signal threads to shut down
            stop_threads.set()

            return jacob
        finally:
            # clean up
            os.remove(util_file)



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="eqsormo",
    version="0.8.0",
    packages=find_packages("src"),
    package_dir={"": "src"},
    install_requires=["statsmodels", "tqdm", "numpy", "pandas", "dill", "scipy",],
    author="Matthew Wigginton Conway",
    author_email="matt@indicatrix.org",
    description="Equilibrium sorting models in Python",
    long_description_content_type="text/markdown",
    long_description=long_description,
    url="https://github.com/mattwigway/eqsormo",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
        "Development Status :: 3 - Alpha",
    ],
)
